###
### Account Services Manager - Subdomains Module
###

# Important module-wide variables
my(@virtual_hosts, $httpd_conf_buffer, %log_config);

# Load the contents of httpd.conf into a buffer, but put all of the vhost definitions into an organized data structure.
sub ParseConfig
{
    open(CONF, $Global_Config{'HTTPD_Config'}[1]);
    my @lines = split(/\n/, join('', <CONF>));
    close(CONF);

    my $reading_vhost = 0;
    my $number_of_vhosts = 0;

    foreach(@lines) {
        if($reading_vhost) {
            if($_ =~ /^\s*\t*\s*<\/VirtualHost>/) {
                $reading_vhost = 0;
                $number_of_vhosts++;
                next;
            }
            else {
                if($_ =~ /^\s*\t*\s*(.*)[\s\t](.*)$/) {
                    $virtual_hosts[$number_of_vhosts]{$1} = $2;
                }
            }
        }
        else {
            if($_ =~ /^\s*\t*\s*<VirtualHost (.*)>/) {
                $reading_vhost = 1;
                $virtual_hosts[$number_of_vhosts]{'VirtualHost'} = $1;
            }
            else {
                $httpd_conf_buffer .= "$_\n";
            }
        }
    }
}

# Write all of the configuration data back to httpd.conf. The only downside to this code is that
# all the virtual hosts will be moved to the end of the file, after all the global directives.
# But is that really such a bad thing, anyway?
sub WriteConfig
{
    open(CONF, ">$Global_Config{'HTTPD_Config'}[1]");
    print CONF $httpd_conf_buffer;
    for(my $i = 0; $i <= $#virtual_hosts; $i++) {
        if($virtual_hosts[$i]{'VirtualHost'}) {
            print CONF "<VirtualHost $virtual_hosts[$i]{'VirtualHost'}>\n";
            foreach(keys %{@virtual_hosts[$i]}) {
                if($_ ne "VirtualHost") {
                    print CONF "$_	$virtual_hosts[$i]{$_}\n";
                }
            }
            print CONF "</VirtualHost>\n\n";
        }
    }
    close(CONF);
}

sub LoadLogConfig
{
    my @tmp_buf;

    for($i = 0; $i < $Global_Config{'LogFile'}[0]; $i++) {
        @tmp_buf = split(/\s/, $Global_Config{'LogFile'}[$i + 1]);
        $log_config{$tmp_buf[0]} = $tmp_buf[1];
    }
}

sub ListSubdomains
{
    print "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\" align=\"center\" width=\"95%\">\n";
    print " <tr>\n";
    print "  <td><u>Subdomain Name</u></td>\n";
    print "  <td><u>Document Root</u></td>\n";
    print "  <td><u>Contact E-mail</u></td>\n";
    print "  <td><u>Modify</u></td>\n";
    print "  <td><u>Delete</u></td>\n";
    print " </tr>\n";

    my $subdomains_printed = 0;

    for(my $i = 0; $i <= $#virtual_hosts; $i++) {
        unless(exists $virtual_hosts[$i]{'ServerAlias'} && $virtual_hosts[$i]{'ServerAlias'} eq $UserConf{'Domain'}) {
            if($virtual_hosts[$i]{'ServerName'} =~ /$UserConf{'Domain'}$/) {
                print " <tr>\n";
                print "  <td><a href=\"http://$virtual_hosts[$i]{'ServerName'}\">$virtual_hosts[$i]{'ServerName'}</a></td>\n";
                print "  <td>$virtual_hosts[$i]{'DocumentRoot'}</td>\n";
                if(exists $virtual_hosts[$i]{'ServerAdmin'}) {
                    print "  <td><a href=\"mailto:$virtual_hosts[$i]{'ServerAdmin'}\">$virtual_hosts[$i]{'ServerAdmin'}</a></td>\n";
                }
                else {
                    print "  <td>N/A</td>\n";
                }
                print "  <td><a href=\"index.cgi?module=subdomains&action=modify&name=$virtual_hosts[$i]{'ServerName'}\"><img src=\"/themes/$Global_Config{'CurrentTheme'}[1]/button.gif\" border=\"0\"></a></td>\n";
                print "  <td><a href=\"index.cgi?module=subdomains&action=delete&name=$virtual_hosts[$i]{'ServerName'}\"><img src=\"/themes/$Global_Config{'CurrentTheme'}[1]/button.gif\" border=\"0\"></a></td>\n";
                print " </tr>\n";
                $subdomains_printed++;
            }
        }
    }

    if($subdomains_printed > 1) {
        print " <tr>\n";
        print "  <td colspan=\"5\" align=\"center\"><i>$subdomains_printed subdomains found.</i></td>\n";
        print " </tr>\n";
    }
    elsif($subdomains_printed == 1) {
        print " <tr>\n";
        print "  <td colspan=\"5\" align=\"center\"><i>1 subdomain found.</i></td>\n";
        print " </tr>\n";
    }
    else {
        print " <tr>\n";
        print "  <td colspan=\"5\" align=\"center\"><i>No subdomains found.</i></td>\n";
        print " </tr>\n";
    }

    print "</table>\n";
    print "<p align=\"center\"><font face=\"verdana, arial\" size=\"2\"><a href=\"index.cgi?module=subdomains&action=add\">Add Subdomain</a> \| <a href=\"index.cgi?module=subdomains\">Refresh List</a></font></p>\n";
}

# Add a subdomain to the user's account.
sub AddSubdomain
{
    my $subdomain;
    if(exists $HTTP_Args{'name'}) {
        $subdomain = lc($HTTP_Args{'name'});
    }
    my $virtualhost  = "$subdomain.$UserConf{'Domain'}";
    my $server_admin;
    if(exists $HTTP_Args{'ServerAdmin'}) {
        $server_admin = $HTTP_Args{'ServerAdmin'};
    }
    my $doc_root;
    if(exists $HTTP_Args{'DocumentRoot'}) {
        $doc_root     = $UserConf{'Home'} . '/' . $HTTP_Args{'DocumentRoot'};
    }

    if(!$subdomain)
    {
        print qq~<form method="post" action="$ENV{'SCRIPT_NAME'}">
<input type="hidden" name="module" value="subdomains">
<input type="hidden" name="action" value="add">
<table border="0" cellpadding="2" cellspacing="2" align="center">
 <tr>
  <td colspan="2" align="center"><h3><font face="verdana, arial">Add Subdomain</font></h3></td>
 </tr>
 <tr>
  <td><font face="verdana, arial" size="2">Name:</font></td>
  <td><input type="text" name="name" value="" size="20" maxchars="20" class="inputbox">.<font face="verdana, arial" size="2">$UserConf{'Domain'}</font></td>
 </tr>
 <tr>
  <td><font face="verdana, arial" size="2">Document Root:</font></td>
  <td><select name="DocumentRoot" class="inputbox">~;

    chdir($UserConf{'Home'});
    my $current_dir = '.';

    opendir(HOME, $current_dir);
    my @dirs = readdir(HOME);
    closedir HOME;
    @dirs = API::filterdirs($current_dir, @dirs);

    ### XXX: Recurion is temporarily disabled. The web server locks up for some reason when more than 52 directories are sent back to the client. (Help!)

    my @tmp_arr;
    foreach(@dirs)
    {
        $_ =~ s/^\.\///;
        $current_dir .= "/$_";
        opendir(SUBDIR, $current_dir);
        @tmp_arr = API::filterdirs($current_dir, readdir(SUBDIR));
        push(@dirs, @tmp_arr);
        closedir(SUBDIR);
        $current_dir = '.';
    }
    undef @tmp_arr;

    foreach(sort @dirs)
    {
        print "<option value=\"$_\">$_</option>\n";
    }

print qq~</select></td>
 </tr>
 <tr>
  <td><font face="verdana, arial" size="2">Contact E-mail:</font></td>
  <td><input type="text" name="ServerAdmin" value="postmaster\@$UserConf{'Domain'}" size="30" class="inputbox"></td>
 </tr>
 <tr>
  <td align="center" colspan="2"><input type="submit" value="Submit"><font size="-1"><br><a href="index.cgi?module=subdomains&action=makedir&oldaction=add">Make a new directory...</a></font></td>
 </tr>
</table>
</form>~;
        return 1;
    }

    # Check for invalid characters.
    foreach(split(//, $subdomain))
    {
        if((!API::is_in_array($_, a..z))&&(!API::is_in_array($_, 0..9))&&($_ ne '.')){
            print "<font face=\"verdana, arial\" size=\"2\">Error: '$subdomain' is not a valid subdomain name. Please use only alphanumeric characters or a .</font>\n";
            return 0;
        }
    }

    # We can't let them add a subdomain 5 million characters long, now can we?
    if(length($subdomain) > 20)    # 20 should be more than enough.
    {
        print "<font face=\"verdana, arial\" size=\"2\">Error: Subdomain name too long.</font>\n";
        return 0;
    }

    # Does the document root exist?
    if(!-e $doc_root || !-d $doc_root)
    {
        print "<font face=\"verdana, arial\" size=\"2\">Error: Document root not found!</font>\n";
        return 0;
    }

    # Make sure they don't exceed the max!
    my $MaxSubs = $UserConf{'MaxSubs'};
    if($MaxSubs ne 'x' && $NumberOfSubs >= $MaxSubs)
    {
        print "<font face=\"verdana, arial\" size=\"2\">Error: Your request to add $virtualhost has been declined because you have reached your limit.</font>\n";
        return 0;
    }

    # Listing the same virtual host twice will make Apache go wild - prevent this.
    for(my $i = 0; $i <= $#virtual_hosts; $i++) {
        if($virtual_hosts[$i]{'ServerName'} eq $virtualhost) {
            print "<font face=\"verdana, arial\" size=\"2\">Error: The requested subdomain already exists.</font>\n";
            return;
        }
    }

    # Add the subdomain to the data hash.
    my $index = $#virtual_hosts + 1;
    $virtual_hosts[$index]{'VirtualHost'}  = ($UserConf{'IPVhosting'} eq 'True') ? $UserConf{'IP'} : $virtualhost;
    $virtual_hosts[$index]{'ServerName'}   = $virtualhost;
    $virtual_hosts[$index]{'DocumentRoot'} = $doc_root;
    $virtual_hosts[$index]{'ServerAdmin'}  = $server_admin;
    $virtual_hosts[$index]{'CustomLog'}    = $log_config{'AccessLog'} . ' combined';
    $virtual_hosts[$index]{'ErrorLog'}     = $log_config{'ErrorLog'};

    # Re-write the configuration file on disk.
    WriteConfig();

    # All done.
    API::LogAction("$virtualhost added; DocumentRoot: $doc_root");
    print "<font face=\"verdana, arial\" size=\"2\">Subdomain $virtualhost successfully added. Changes may take up to 5 minutes to activate.</font>\n";
}

# Delete a subdomain.
sub DeleteSubdomain
{
    my $virtualhost = shift();

    for(my $i = 0; $i < $#virtual_hosts; $i++) {
        if($virtual_hosts[$i]{'ServerName'} eq $virtualhost) {
            foreach(keys %{@virtual_hosts[$i]}) {
                delete $virtual_hosts[$i]{$_};
            }
        }
    }
    WriteConfig();

    API::LogAction("$virtualhost deleted.");
    print "<font face=\"verdana, arial\" size=\"2\">$virtualhost deleted successfully. Changes may take up to 10 minutes to activate.</font>\n";
}

# Modify a subdomain
sub ModifySubdomain
{
    my $virtualhost  = shift();
    my $doc_root     = $HTTP_Args{'DocumentRoot'};
    my $server_admin = $HTTP_Args{'ServerAdmin'};

    if($virtualhost !~ /$UserConf{'Domain'}$/) {
        print "<font face=\"verdana, arial\" size=\"2\">Error: You are not authorized to modify this subdomain.</font>\n";
        return;
    }

    my %current;
    for(my $i = 0; $i <= $#virtual_hosts; $i++)
    {
        if($virtual_hosts[$i]{'ServerName'} eq $virtualhost)
        {
            next if($virtual_hosts[$i]{'ServerAlias'} eq $UserConf{'Domain'});    # Skip the main virtualhost, that's not a subdomain.
            $current{'DocumentRoot'} = $virtual_hosts[$i]{'DocumentRoot'};
            $current{'ServerAdmin'}  = $virtual_hosts[$i]{'ServerAdmin'};
            $current{'DocumentRoot'} =~ s/^$UserConf{'Home'}\///i;
            last;
        }
    }

    if(!$doc_root || !$server_admin){
        print qq~<form method="post" action="$ENV{'SCRIPT_NAME'}">
<input type="hidden" name="module" value="subdomains">
<input type="hidden" name="action" value="modify">
<input type="hidden" name="name" value="$virtualhost">
<table border="0" cellpadding="1" cellspacing="0" align="center" width="400">
 <tr>
  <td colspan="2" align="center"><font face=\"verdana, arial\" size="+1"><b><u>Modify Subdomain</u></b></font></td>
 </tr>
 <tr>
  <td colspan="2" align="center"><font face=\"verdana, arial\" size=\"2\">Modifiying subdomain $virtualhost</font></td>
 </tr>
 <tr>
  <td><font face=\"verdana, arial\" size=\"2\">Document root:</font></td>
  <td><select name="DocumentRoot" class="inputbox">
~;

    chdir($UserConf{'Home'});
    my $current_dir = '.';

    opendir(HOME, $current_dir);
    my @dirs = readdir(HOME);
    closedir HOME;
    @dirs = API::filterdirs($current_dir, @dirs);

    my @tmp_arr;
    foreach(@dirs)
    {
        $_ =~ s/^\.\///;
        $current_dir .= "/$_";
        opendir(SUBDIR, $current_dir);
        @tmp_arr = API::filterdirs($current_dir, readdir(SUBDIR));
        @dirs = (@dirs, @tmp_arr);
        closedir(SUBDIR);
        $current_dir = '.';
    }
    undef @tmp_arr;

    $doc_root =  $current{'DocumentRoot'};
    $doc_root =~ s/^$UserConf{'Home'}//i;

    foreach(sort @dirs)
    {
        if($_ eq $doc_root){
            print "   <option value=\"$_\" selected>$_</option>\n";
        }else{
            print "   <option value=\"$_\">$_</option>\n";
        }
    }

    print qq~</td>
 </tr>
 <tr>
  <td><font face=\"verdana, arial\" size=\"2\">Contact email:</font></td>
  <td><input type="text" name="ServerAdmin" value="$current{'ServerAdmin'}" class="inputbox"></td>
 </tr>
 <tr>
  <td colspan="2" align="center"><input type="submit" value="Submit Changes"></td>
 </tr>
</table>
</form>~;
    return 1;
    }

    if(!-e "$UserConf{'Home'}/$doc_root" || !-d "$UserConf{'Home'}/$doc_root")
    {
        print "<font face=\"verdana, arial\" size=\"2\">Error: Document root not found!</font>\n";
        return 0;
    }

    my $changes = 0;
    for(my $i = 0; $i <= $#virtual_hosts; $i++) {
        if($virtual_hosts[$i]{'ServerName'} eq $virtualhost) {
            next if($virtual_hosts[$i]{'ServerAlias'} eq $UserConf{'Domain'});    # Skip the main virtualhost, that's not a subdomain.

            if($virtual_hosts[$i]{'DocumentRoot'} ne "$UserConf{'Home'}/$doc_root") {
                $virtual_hosts[$i]{'DocumentRoot'} = "$UserConf{'Home'}/$doc_root";
                $changes++;
            }
            if($virtual_hosts[$i]{'ServerAdmin'} ne $server_admin) {
                $virtual_hosts[$i]{'ServerAdmin'} = $server_admin;
                $changes++;
            }
        }
    }

    if($changes > 1) {
        WriteConfig();
        print "<font face=\"verdana, arial\" size=\"2\">$changes changes were made to your subdomain. Changes will take effect within 5 minutes.</font>\n";
    }
    elsif($changes == 1) {
        WriteConfig();
        print "<font face=\"verdana, arial\" size=\"2\">The change to your subdomain will take effect within 5 minutes.</font>\n";
    }
    else {
        print "<font face=\"verdana, arial\" size=\"2\">Press your browser's back button and make some changes before pressing the Update button.</font>\n";
    }
}

# Make a new directory.
sub MakeDir
{
    my $dir = $_[0];

    if(!$dir) {
        print "<h1 align=\"center\"><u>Make New Directory</u></h1>\n";
        print "<p align=\"center\">\n";
        print "<form method=\"get\" action=\"index.cgi\">\n";
        print "<input type=\"hidden\" name=\"module\" value=\"subdomains\">\n";
        print "<input type=\"hidden\" name=\"action\" value=\"makedir\">\n";
        print "<input type=\"hidden\" name=\"oldaction\" value=\"$HTTP_Args{'oldaction'}\">\n";
        print "<font face=\"verdana, arial\" size=\"2\">Name of folder to create: </font><input type=\"text\" name=\"name\" value=\"\" max=\"40\"><br>\n";
        print "<input type=\"submit\" value=\"Submit\">\n";
        print "</form>\n";
        print "</p>\n";
    }
    else {
        if($dir =~ /^\// || $dir =~ /\.\./) {
            print "<b><font face=\"verdana, arial\" size=\"2\" color=\"#FF0000\">Invalid directory. You may not use .. in the folder name, nor may it begin with a /.</font></b>\n";
        }
        else {
            if(mkdir("$UserConf{'Home'}/$dir", 0755)) {
                chown((getpwnam($AUTH::User))[2, 3], "$UserConf{'Home'}/$dir");
                print "<font face=\"verdana, arial\" size=\"2\">The directory $UserConf{'Home'}/$dir was successfully created.</font>\n";
                API::LogAction("User $AUTH::User created directory $UserConf{'Home'}/$dir");
            }
            else {
                print "<font face=\"verdana, arial\" size=\"2\">Failed to create directory. The operating system returned the following error: $!</font>\n";
            }
        }
    }
}

# Print a link at the bottom to make life a little easier.
sub PrintBackLink
{
        print "<p align=\"center\"><font face=\"verdana, arial\" size=\"2\"><a href=\"index.cgi?module=subdomains&action=$HTTP_Args{'oldaction'}\">Back...</a></font></p>\n";
}

sub Main
{
    ParseConfig();
    LoadLogConfig();
    my $action = $HTTP_Args{'action'};

    # Do whatever it is they're trying to do *before* we display the list.
    if($action eq 'add')
    {
        AddSubdomain();
        PrintBackLink();
    }
    elsif($action eq 'delete')
    {
        DeleteSubdomain($HTTP_Args{'name'});
    }
    elsif($action eq 'modify')
    {
        ModifySubdomain($HTTP_Args{'name'});
        PrintBackLink();
    }
    elsif($action eq 'makedir')
    {
        MakeDir($HTTP_Args{'name'});
        PrintBackLink();
    }
    else {
        ListSubdomains();
    }
}

1;
