/* 
 * Account Services Manager - Adduser
 *     Copyright (C) 2001 Brandon Ramirez. All rights reserved.
 *
 * For copying information, please see the file docs/licsence.txt
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <crypt.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/types.h>

#include "config.h"
#include "salt.h"

//
// Program-specific memory management.
//
static char *username;
static char *password;
static char *home_dir;
static char *ip_addr;
static char *domain;
static int  set_pw;

static void FreeMemory()
{
    delete username;
    if(set_pw)
        delete password;
    delete home_dir;
    delete ip_addr;
    delete domain;
}

void die(int code)
{
    FreeMemory();
    if(&code == NULL)
        code = 0;
    exit(code);
}

/*
 * Main function...
 */
int main(int argc, char *argv[])
{
    char *pass_buff;
    char *directory;
    mode_t mask;
    char *crypted_password;
    dirent *direntry;
    DIR  *d;
    FILE *f;

    // Free up memory if they decide to quit ahead of time.
    signal(SIGTERM, die);
    signal(SIGQUIT, die);

    // Do we set the password for this user in their profile?
    if(argc == 2 && (strcmp(argv[1], "-p") == 0))
        set_pw = 1;
    else
        set_pw = 0;

    //
    // Step 1) Username
    //
    username = new char[MAX_USERNAME + 1];
    printf("Username: ");
    fgets(username, MAX_USERNAME, stdin);
    username[strlen(username) - 1] = '\0';

    //
    // Step 2) Password
    //
    if(set_pw) {
        pass_buff = getpass("Password: ");
        password = new char[MAX_PASSWORD + 1];
        strcpy(password, pass_buff);
        memset(pass_buff, 0, strlen(pass_buff));
        delete pass_buff;
    }

    //
    // Step 3) Home Directory
    //
    home_dir = new char[MAX_HOME_DIR + 1];
    printf("Home Directory: ");
    fgets(home_dir, MAX_HOME_DIR, stdin);
    home_dir[strlen(home_dir) - 1] = '\0';

    //
    // Step 4) IP Address
    //
    ip_addr = new char[MAX_IP_ADDR + 1];
    printf("IP Address: ");
    fgets(ip_addr, MAX_IP_ADDR, stdin);
    ip_addr[strlen(ip_addr) - 1] = '\0';

    //
    // Step 5) Domain Name
    //
    domain = new char[MAX_DOMAIN + 1];
    printf("Domain Name: ");
    fgets(domain, MAX_DOMAIN, stdin);
    domain[strlen(domain) - 1] = '\0';

    //
    // Step 6) Do something with this data!
    //

    // Change our current directory.
    directory = new char[strlen(BASE_DIR) + 7];
    sprintf(directory, "%s/users", BASE_DIR);
    chdir(directory);

    // Does this account already exist?
    d = opendir(".");
    while((direntry = readdir(d)) != NULL)
    {
        if(strcmp(direntry->d_name, username) == 0)
        {
            fprintf(stderr, "Error: User %s already exists!\n", username);
            die(1);
        }
    }
    closedir(d);

    // Open the file for writing.
    mask = umask(077);
    f = fopen(username, "w");
    umask(mask);
    if(f == NULL)
    {
        fprintf(stderr, "Error: Failed to create file %s/%s!\n", directory, username);
        die(2);
    }

    // Encrypt the password.
    if(set_pw)
        crypted_password = crypt(password, crypt_make_salt());

    // This is the part that actually writes the stuff out into the file.
    if(set_pw)
        fprintf(f, "Passwd: %s\n", crypted_password);
    fprintf(f, "Home: %s\n", home_dir);
    fprintf(f, "IP: %s\n", ip_addr);
    fprintf(f, "Domain: %s\n", domain);

    // Close the file descriptor.
    fclose(f);

    // Tell them it was a success.
    printf("\nUser %s successfully added!\n", username);

    // Free all the memory we allocated.
    FreeMemory();

    return(0);
}
